# Copyright (c) 2014-2019 Sine Nomine Associates
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THE SOFTWARE IS PROVIDED 'AS IS' AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

"""OpenAFS utilities"""

from __future__ import print_function
import sys
from afsutil.cli import subcommand, argument, usage, dispatch
import afsutil.system
import afsutil.build

@subcommand()
def version(**args):
    """Print version"""
    from afsutil import __version__
    print("afsutil version", __version__)
    return 0

@subcommand()
def help(**args):
    "Print usage"
    return usage()

@subcommand(
    argument('--creds', help='path or url of repo creds (solaris only)',
                        default='/root/creds'),
    argument('-n', '--dry-run', help='do not make changes',
                                dest='dryrun', action='store_true'),
    argument('--skip-headers', help='do not install kernel headers',
                               action='store_true'),
    argument('--skip-solarisstudio', help='do not install solaris studio (solaris only)',
                                     action='store_true'),
    )
def getdeps(**args):
    "Install build dependencies"
    from afsutil.getdeps import getdeps
    return getdeps(**args)

@subcommand(
    argument('--fix-hosts', help='fix /etc/hosts file by replacing loopback entry.',
                            action='store_true')
    )
def check(**args):
    "Check hostname"
    from afsutil.check import check
    return check(**args)

@subcommand(
    argument('--chdir', help='change to directory'),
    argument('--cf', help='configure options', default=afsutil.build.cfopts()),
    argument('--xcf', help='extra configure options', nargs='*', action='append', default=[]),
    argument('--target', help='make target', default='all'),
    argument('--clean', help='run git clean (requires: git config afsutil.clean true)',
                        default=False, action='store_true'),
    argument('-j', '--jobs', help='parallel build jobs', default=afsutil.system.nproc()),
    argument('--srcdir', help='source code directory', default='.'),
    argument('--tarball', help='path and file name of destination tarball when Transarc-style is enabled.'),
    argument('-p', '--paths', help="command paths: <cmd>=<path-to-cmd>",
                              nargs='+', action='append', default=[]),
    )
def build(**args):
    "Build binaries"
    from afsutil.build import build
    return build(**args)

@subcommand(
    argument('--chdir', help='change to directory'),
    requires_root=True,
    )
def reload(**args):
    "Reload the kernel module from the build tree"
    from afsutil.modreload import modreload
    return modreload(**args)

@subcommand(
    argument('--chdir', help='change to directory', metavar='<path>'),
    argument('--mock', help='build in a mock chroot',
                           action='store_true', default='default'),
    argument('--chroot', help='mock chroot configuration as defined in /etc/mock/<chroot>.cfg (--mock only)',
                           metavar='<chroot>', default='default'),
    argument('--dstdir', help='where to place rpms (--mock only)',
                           metavar='<dstdir>'),
    argument('--clobber', help='rebuild and overwrite existing kmods (--mock only)',
                           action='store_true'),
    argument('--no-clean', help='do not clean after build (--mock only)', dest='autoclean', action='store_false'),
    argument('--build', help='what to build: all, sources, srpm, userspace, kmods', metavar='<target>', default='all',
                        choices=['all','sources','srpm','userspace','kmods']),
    argument('--csdb', help='CellServDB file path or url', metavar='<csdb>'),
    argument('--spec', help='spec file path', metavar='<spec>'),
    argument('--srpm', help='prebuilt srpm file path', metavar='<spec>'),
    argument('--version', help='target version number', metavar='<afsversion>'),
    argument('--arch', help='target architecture, e.g. x86_64', metavar='<arch>'),
    argument('-k','--kernel-version', help='linux kernel versions (default: check)',
                           metavar='<kernel-version>', dest='kversions',
                           nargs='+', action='append', default=[]),
    argument('--list-kversions', help='list linux kernel versions then exit', action='store_true'),
    argument('--with', help='build with optional feature (e.g. --with kauth)',
                       dest='with_', metavar='<feature>',
                       nargs='+', action='append', default=[]),
    argument('--without', help='build without optional feature (e.g. --without kauth)',
                       metavar='<feature>',
                       nargs='+', action='append', default=[]),
    )
def package(**args):
    "Build packages"
    from afsutil.package import package
    return package(**args)

@subcommand(
    argument('--chdir', help='change to directory'),
    argument('--dist', help='distribution type',
                       choices=['transarc', 'rpm', 'yum'], default='transarc'),
    argument('--dir', help='distribution directory'),
    argument('--components', help='components to install',
                             metavar='<name>',
                             choices=['client', 'server'],
                             nargs='+', default=['client', 'server']),
    argument('--cell', help='cell name', default='localcell'),
    argument('--hosts', help='cell service db hosts', nargs='+', default=[]),
    argument('--realm', help='realm name'),
    argument('--csdb', help='path to CellServDB.dist file for client'),
    argument('--force', help='overwrite existing files', action='store_true'),
    argument('-o', '--options', help="command line args: <name>=<value>",
                                action='append', nargs='+', default=[]),
    argument('--pre', help='pre-install command', dest='pre_install'),
    argument('--post', help='post-install command', dest='post_install'),
    requires_root=True,
    )
def install(**args):
    "Install binaries"
    from afsutil.install import install
    return install(**args)

@subcommand(
    argument('--dist', help='distribution type',
                       choices=['transarc', 'rpm', 'yum'], default='transarc'),
    argument('--components', help='components to remove',
                             metavar='<name>', nargs='+',
                             choices=['client', 'server'],
                             default=['client', 'server']),
    argument('--purge', help='remove config and data too', action='store_true'),
    argument('--pre', help='pre-remove command', dest='pre_remove'),
    argument('--post', help='post-remove command', dest='post_remove'),
    requires_root=True,
    )
def remove(**args):
    "Remove binaries"
    from afsutil.install import remove
    return remove(**args)

@subcommand(
    argument('components', help='Services to start',
                           metavar='<name>', nargs='*',
                           default=['client','server']),
    requires_root=True,
    )
def start(**args):
    "Start AFS services"
    from afsutil.service import start
    return start(**args)

@subcommand(
    argument('components', help='services to stop',
                           metavar='<name>', nargs='*',
                           default=['client','server']),
    requires_root=True,
    )
def stop(**args):
    "Stop AFS services"
    from afsutil.service import stop
    return stop(**args)

@subcommand(
    argument('--keytab', help='keytab file to be created', default='/tmp/afs.keytab'),
    argument('--cell', help='cell name', default='localcell'),
    argument('--realm', help='realm name'),
    argument('--enctype', help='encryption type', default='aes256-cts-hmac-sha1-96'),
    argument('--secret', help='passphrase'),
    )
def ktcreate(**args):
    "Create a fake keytab"
    from afsutil.keytab import create
    return create(**args)

@subcommand(
    argument('--keytab', help='keytab file to be destroyed', default='/tmp/afs.keytab'),
    argument('--force', help='ignore errors', action='store_true'),
    )
def ktdestroy(**args):
    "Destroy a keytab"
    from afsutil.keytab import destroy
    return destroy(**args)

@subcommand(
    argument('--keytab', help="keytab file", default="/tmp/afs.keytab"),
    argument('--cell', help="cell name"),
    argument('--realm', help="realm name" ),
    argument('--confdir', help="server config directory", default="/usr/afs/etc"),
    argument('--format', help="key file format",
                         choices=['detect', 'transarc', 'rxkad-k5', 'extended'],
                         default='detect', dest='kformat'),
    argument('-n', '--dry-run', help="do not make changes",
                                dest='dryrun', action='store_true'),
    argument('-p', '--paths', help="command paths: <cmd>=<path-to-cmd>",
                                nargs='+', action='append', default=[]),
    requires_root=True,
    )
def ktsetkey(**args):
    "Add a service key from a keytab file"
    from afsutil.keytab import setkey
    return setkey(**args)

@subcommand(
    argument('--akimpersonate', help="print a ticket for the service key in the keytab", action='store_true'),
    argument('--keytab', help="keytab file", default="/tmp/afs.keytab"),
    argument('--user', help="user name", default="admin"),
    argument('--cell', help="cell name", default="localcell"),
    argument('--realm', help="realm name", default="LOCALCELL"),
    argument('-p', '--paths', help="command paths: <cmd>=<path-to-cmd>",
                                nargs='+', action='append', default=[]),
    )
def ktlogin(**args):
    "Obtain a token with a keytab"
    from afsutil.cell import login
    return login(**args)

@subcommand(
    argument('--cell', help="cell name", default='localcell'),
    argument('--realm', help="realm name"),
    argument('--admin', help="admin username", default='admin'),
    argument('--db', help="cell database hosts", nargs='+', default=[]),
    argument('--fs', help="cell fileserver hosts", nargs='+', default=[]),
    argument('-o', '--options', help="command line args: <[hostname:]name>=<value>",
                                nargs='+', action='append', default=[]),
    argument('-p', '--paths', help="command paths: <cmd>=<path-to-cmd>",
                                nargs='+', action='append', default=[]),
    requires_root=True,
    )
def newcell(**args):
    "Setup a new cell"
    from afsutil.cell import newcell
    return newcell(**args)

@subcommand(
    argument('--cell', help="cell name", default='localcell'),
    argument('--realm', help="realm name"),
    argument('--akimpersonate', help="print a ticket for admin user", action='store_true'),
    argument('--keytab', help="keytab file", default="/tmp/afs.keytab"),
    argument('--admin', help="admin username", default='admin'),
    argument('--fs', help="cell fileserver hosts", nargs='+', default=[]),
    argument('--top', help="top level volumes", nargs='+', default=[]),
    argument('--aklog', help="path to aklog program"),
    argument('--kinit', help="path to kinit program"),
    argument('-o', '--options', help="command line args: <name>=<value>",
                                action='append', nargs='+', default=[]),
    argument('-p', '--paths', help="command paths: <cmd>=<path-to-cmd>",
                                action='append', nargs='+', default=[]),
    )
def mtroot(**args):
    "Mount root volumes in a new cell"
    from afsutil.cell import mtroot
    return mtroot(**args)

@subcommand(
    argument('hostname', help="fileserver hostname"),
    argument('--keytab', help="keytab file", default="/tmp/afs.keytab"),
    argument('-o', '--options', help="command line args; <name>=<value>",
                                nargs='+', action='append', default=[]),
    argument('-p', '--paths', help="command paths: <cmd>=<path-to-cmd>",
                                action='append', nargs='+', default=[]),
    requires_root=True,
    )
def addfs(**args):
    "Add a new fileserver to a cell"
    from afsutil.cell import addfs
    return addfs(**args)

def main():
    return dispatch()

if __name__ == '__main__':
    sys.exit(main())



# Copyright (c) 2014-2015 Sine Nomine Associates
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THE SOFTWARE IS PROVIDED 'AS IS' AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
#

import unittest
import tempfile
import subprocess
import os

import afsutil.cmd
import afsutil.system
from afsutil.keytab import Keytab, _split_principal
from afsutil.keytab import _check_for_extended_keyfile_support

have_klist = afsutil.system.which("klist")
asetkey_new= os.path.join(os.path.dirname(os.path.abspath(__file__)), "data", "asetkey-new")
asetkey_old= os.path.join(os.path.dirname(os.path.abspath(__file__)), "data", "asetkey-old")

class KeytabTest(unittest.TestCase):
    """Requires kinit and a pregenerated keytab."""

    def __init__(self, *args, **kwargs):
        # Test with keytab previously created with kadmin.
        unittest.TestCase.__init__(self, *args, **kwargs)
        self.testfile = os.path.join(os.path.dirname(os.path.abspath(__file__)), "data", "example.keytab")
        assert os.path.exists(self.testfile)

    def test_split_principal(self):
        self.assertRaises(ValueError, _split_principal, "missing_realm")
        c,r = _split_principal("myname@MYREALM")
        self.assertEqual(c, ["myname"])
        self.assertEqual(r, "MYREALM")
        c,r = _split_principal("afs/mycell@MYREALM")
        self.assertEqual(c, ["afs", "mycell"])
        self.assertEqual(r, "MYREALM")

    def test_read_keytab(self):
        k = Keytab.load(self.testfile)
        self.assertIsNotNone(k)
        self.assertEquals(len(k.entries), 16)
        self.assertEquals(k.entries[0]['realm'], "LOCALREALM")
        self.assertEquals(k.entries[0]['principal'], "afs/robotest@LOCALREALM")
        self.assertEquals(k.entries[0]['kvno'], 10)
        self.assertEquals(k.entries[0]['eno'], 18)
        self.assertEquals(k.entries[0]['enctype'], 'aes256-cts-hmac-sha1-96')

    @unittest.skipUnless(have_klist, "missing klist")
    def test_create_empty_file(self):
        fd,filename = tempfile.mkstemp()
        try:
            k = Keytab()
            k.write(filename)
            self.assertEqual(k.entries, [])
            out = subprocess.check_output(["klist", "-e", "-k", "-t", filename]).splitlines()
            self.assertEquals(len(out), 3)
            self.assertTrue(out[0].startswith("Keytab name"))
            self.assertTrue(out[1].startswith("KVNO Timestamp"))
            self.assertTrue(out[2].startswith("----"))
        finally:
            os.close(fd)
            os.remove(filename)

    @unittest.skipUnless(have_klist, "missing klist")
    def test_create_fake_aes_key(self):
        fd,filename = tempfile.mkstemp()
        try:
            k = Keytab()
            k.add_fake_key("afs/localcell@LOCALCELL")
            k.write(filename)
            out = subprocess.check_output(["klist", "-e", "-k", "-t", filename]).splitlines()
            self.assertEquals(len(out), 4)
            self.assertTrue(out[0].startswith("Keytab name"))
            self.assertTrue(out[1].startswith("KVNO Timestamp"))
            self.assertTrue(out[2].startswith("----"))
            self.assertTrue("afs/localcell@LOCALCELL" in out[3])
            self.assertTrue("aes256-cts-hmac-sha1-96" in out[3])
        finally:
            os.close(fd)
            os.remove(filename)

    @unittest.skipUnless(have_klist, "missing klist")
    def test_create_fake_des_key(self):
        fd,filename = tempfile.mkstemp()
        try:
            k = Keytab()
            k.add_fake_key("afs/localcell@LOCALCELL", enctype='des-cbc-crc')
            k.write(filename)
            out = subprocess.check_output(["klist", "-e", "-k", "-t", filename]).splitlines()
            self.assertEquals(len(out), 4)
            self.assertTrue(out[0].startswith("Keytab name"))
            self.assertTrue(out[1].startswith("KVNO Timestamp"))
            self.assertTrue(out[2].startswith("----"))
            self.assertTrue("afs/localcell@LOCALCELL" in out[3])
            self.assertTrue("des-cbc-crc" in out[3])
        finally:
            os.close(fd)
            os.remove(filename)


    def test_get_kvno(self):
        k = Keytab()
        k.read(self.testfile)
        self.assertEquals(k.get_kvno("afs/robotest@LOCALREALM"), 10)
        self.assertEquals(k.get_kvno("robotest@LOCALREALM"), 19)
        self.assertEquals(k.get_kvno("bogus@BOGUS"), None)

    def test_get_entries(self):
        k = Keytab()
        k.read(self.testfile)
        entries = k.get_entries("afs/robotest@LOCALREALM")
        self.assertIsNotNone(entries)
        self.assertEquals(len(entries), 4)
        self.assertEquals(entries[0]['kvno'], 10)

    def test_get_entries_for_kvno(self):
        k = Keytab()
        k.read(self.testfile)
        entries = k.get_entries("robotest@LOCALREALM", 18) # not the max kvno
        self.assertIsNotNone(entries)
        self.assertEquals(len(entries), 4)

    def test_is_des(self):
        k = Keytab()
        k.read(self.testfile)
        self.assertTrue(k.is_des("afs/plugh@LOCALREALM"))
        self.assertFalse(k.is_des("afs/xyzzy@LOCALREALM"))
        # The following has a mix of des and non-des.
        self.assertRaises(AssertionError, Keytab.is_des, k, "afs/robotest@LOCALREALM")
        # No keys for the following.
        self.assertRaises(AssertionError, Keytab.is_des, k, "afs/bogus@LOCALREALM")

    def test_service_principals_0(self):
        k = Keytab()
        self.assertEquals(len(k.service_principals()), 0)

    def test_service_principals_1(self):
        k = Keytab()
        k.add_fake_key("afs/foobar@FOOBAR")
        self.assertEquals(len(k.service_principals()), 1)

    def test_service_principals_n(self):
        k = Keytab.load(self.testfile)
        self.assertItemsEqual(k.service_principals(),
            ['afs/xyzzy@LOCALREALM', 'afs/robotest@LOCALREALM', 'afs/plugh@LOCALREALM'])

    @unittest.skipIf(not os.path.exists(asetkey_old), "missing test old asetkey")
    def test_check_for_no_extended_keyfile_support(self):
        origprog = afsutil.cmd.ASETKEY
        afsutil.cmd.ASETKEY = asetkey_old
        self.assertFalse(_check_for_extended_keyfile_support())
        afsutil.cmd.ASETKEY = origprog

    @unittest.skipIf(not os.path.exists(asetkey_new), "missing test new asetkey")
    def test_check_for_extended_keyfile_support(self):
        origprog = afsutil.cmd.ASETKEY
        afsutil.cmd.ASETKEY = asetkey_new
        self.assertTrue(_check_for_extended_keyfile_support())
        afsutil.cmd.ASETKEY = origprog

    @unittest.skipIf(not os.path.exists(asetkey_new), "missing test new asetkey")
    def test_guess_key_format(self):
        origprog = afsutil.cmd.ASETKEY
        afsutil.cmd.ASETKEY = asetkey_new
        k = Keytab()
        principal = "afs/xyzzy@LOCALREALM"
        k.add_fake_key(principal)
        self.assertEquals(k._guess_key_format(principal), 'extended')
        afsutil.cmd.ASETKEY = origprog

if __name__ == "__main__":
     unittest.main()




# Copyright (c) 2014-2017 Sine Nomine Associates
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THE SOFTWARE IS PROVIDED 'AS IS' AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

"""Reload the client after rebuilding from source (developer tool)"""

import logging
import os
import re

from afsutil.system import file_should_exist
import afsutil.service
import afsutil.transarc

logger = logging.getLogger(__name__)

def _linux_kernel_packaging():
    """Returns true if built with --with-linux-kernel-packaging."""
    value = False
    with open('src/config/Makefile.config', 'r') as f:
        for line in f.readlines():
            m = re.search(r'^LINUX_KERNEL_PACKAGING = (\S+)', line)
            if m:
                value = (m.group(1) == 'yes')
                break
    return value

def _linux_libafs_name():
    """Get the libafs filename."""
    sysname, nodename, release, version, machine = os.uname()
    if _linux_kernel_packaging():
        libafs = 'openafs.ko'
    else:
        mp = '.mp' if 'SMP' in version else ''
        libafs = "libafs-%s%s.ko" % (release, mp)
    return libafs

def _linux_modload_name():
    """Get the modload directory name."""
    sysname, nodename, release, version, machine = os.uname()
    if _linux_kernel_packaging():
        modload = "MODLOAD-%s-SP" % (release)
    else:
        mp = 'MP' if 'SMP' in version else 'SP'
        modload = "MODLOAD-%s-%s" % (release, mp)
    return modload

def _kmod():
    """Get the kernel module file name in the build tree."""
    sysname = os.uname()[0]
    if sysname == 'Linux':
        libafs = _linux_libafs_name()
        modload = _linux_modload_name()
        kmod = os.path.abspath("src/libafs/%s/%s" % (modload, libafs))
    elif sysname == 'SunOS':
        kmod = os.path.abspath("src/libafs/MODLOAD64/libafs.o")
    else:
        raise AssertionError("Unsupported operating system: %s" % (sysname))
    return kmod

def _client_setup():
    """Get the client setup object for this os."""
    uname = os.uname()[0]
    if uname == "Linux":
        cs = afsutil.transarc.LinuxClientSetup()
    elif uname == "SunOS":
        cs = afsutil.transarc.SolarisClientSetup()
    else:
        raise AssertionError("Unsupported operating system: %s" % (uname))
    return cs

def modreload(**kwargs):
    """Reload the kernel module and restart the cache manager after a build.

    This is a helper for testing and debugging the afs kernel module and afsd
    without resorting to a full reinstallation after every code change. The cache
    manager must be installed at least once, then can be reloaded after each code
    change and build with this function.

    This should be run as root in the top level source directory. At this time,
    only transarc-style installations are supported by modreload.

    A typical workflow, using the afsutil front-end script:

        $ sudo afsutil install --dist transarc [options]
        $ sudo afsutil start client
        ... test/debug/etc ...
        ... edit files in src/afs ...

        $ make libafs
        $ sudo afsutil reload
        ... test/debug/etc ...
        ... edit files in src/afs ...

        $ make libafs
        $ sudo afsutil reload
        ... repeat ...
    """
    cs = _client_setup()
    afsd = os.path.abspath('src/afsd/afsd')
    kmod = _kmod()
    file_should_exist(afsd)        # Check before stopping.
    file_should_exist(kmod)
    afsutil.service.stop(components=['client'])
    cs.install_afsd(afsd)
    cs.install_kmod(kmod)
    afsutil.service.start(components=['client'])



# Copyright (c) 2014-2017 Sine Nomine Associates
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THE SOFTWARE IS PROVIDED 'AS IS' AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
#

"""Utility to set the AFS service key."""

import hashlib
import os
import struct
import time
import logging
import shutil

from afsutil.system import CommandFailed, mkdirp
from afsutil.transarc import AFS_CONF_DIR
from afsutil.cmd import asetkey, aklog

logger = logging.getLogger(__name__)

KFORMATS = ['transarc', 'rxkad-k5', 'extended']
DEFAULT_KEYTAB = "/tmp/afs.keytab"
DEFAULT_USER = "admin"

KRB_ENCTYPES = {
    1: 'des-cbc-crc',
    2: 'des-cbc-md4',
    3: 'des-cbc-md5',
    5: 'des3-cbc-md5',
    7: 'des3-cbc-sha1',
    9: 'dsaWithSHA1-CmsOID',
    10: 'md5WithRSAEncryption-CmsOID',
    11: 'sha1WithRSAEncryption-CmsOID',
    12: 'rc2CBC-EnvOID',
    13: 'rsaEncryption-EnvOID',
    14: 'rsaES-OAEP-ENV-OID',
    15: 'des-ede3-cbc-Env-OID',
    16: 'des3-cbc-sha1-kd',
    17: 'aes128-cts-hmac-sha1-96',
    18: 'aes256-cts-hmac-sha1-96',
    23: 'rc4-hmac',
    24: 'rc4-hmac-exp',
    25: 'camellia128-cts-cmac',
    26: 'camellia256-cts-cmac',
    65: 'subkey-keymaterial',
}

# Helpers

def _split_principal(principal):
    """Split a fully qualified principal string into (components, realm)."""
    if not '@' in principal:
        raise ValueError("Missing realm name in principal")
    names,realm = principal.split('@')
    comps = names.split('/')
    return (comps, realm)

def _lookup_eno(enctype):
    """Lookup the eno (number) from the enctype (string)."""
    enos = {}
    for k,v in KRB_ENCTYPES.items():
        assert not v in enos
        enos[v] = k
    try:
        eno = enos[enctype]
    except KeyError:
        raise ValueError("Unsupported enctype '%s'." % enctype)
    return eno

def _is_des_eno(eno):
    """Return true if the eno is for a DES enctype."""
    return (eno in (1,2,3,15))

def _check_for_extended_keyfile_support():
    """Returns true if asetkey supports the new extended key files.

    Runs asetkey without options to elicit the usage message, then
    checks for the new 'add' subcommand."""
    logger.debug("Checking if asetkey supports extended key files.")
    usage = None
    try:
        usage = asetkey(quiet=True)
    except CommandFailed as e:
        usage = e.out
    logger.debug("asetkey usage: %s", usage)
    if usage is None or not "usage" in usage:
        raise AssertionError("Failed to get asetkey usage.")
    new_asetkey = "add <type> <kvno> <subtype> <keyfile> <princ>" in usage
    if new_asetkey:
        logger.info("asetkey supports extended key format.")
    else:
        logger.info("asetkey does not support extended key format.")
    return new_asetkey

class KeytabException(Exception):
    pass

class InvalidKeytab(KeytabException):
    pass

class NotFoundKeytab(KeytabException):
    pass

class Keytab(object):
    """Utility to support testing with keytabs and akimpersonate.

    Examples:
        from afsutil.keytab import Keytab
        principal = "afs/mycell@MYREALM"

        # Create a fake keytab file for testing with akimpersonate.
        k = Keytab()
        k.add_fake_key(principal)
        k.write("path/to/my/fake.keytab")

        # Read an existing keytab file and find entries.
        k = Keytab().load("path/to/my/file.keytab")
        kvno = k.get_kvno(principal)
        for e in k.get_entries(principal):
            print e['kvno'], e['principal'], e['enctype']
        if k.is_des(principal):
            print "DES service key present"

        # Create an empty keytab file.
        # The created file will have the current user permissions
        # and can be updated with sudo kadmin.local.
        k = Keytab()
        k.write("path/to/my/empty.keytab")


    """
    _KRB_KEYTAB_MAGIC_OLD = 0x0501
    _KRB_KEYTAB_MAGIC = 0x0502
    _KRB_NT_PRINCIPAL = 1

    # The following C-like structure definitions illustrate the MIT keytab
    # file format. All values are in network byte order. All text is ASCII.
    #
    #   keytab {
    #       uint16_t file_format_version;                    /* 0x502 */
    #       keytab_entry entries[*];
    #   };
    #   keytab_entry {
    #       int32_t size;
    #       uint16_t num_components;    /* sub 1 if version 0x501 */
    #       counted_octet_string realm;
    #       counted_octet_string components[num_components];
    #       uint32_t name_type;   /* not present if version 0x501 */
    #       uint32_t timestamp;
    #       uint8_t vno8;
    #       keyblock key;
    #       uint32_t vno; /* only present if >= 4 bytes left in entry */
    #   };
    #   counted_octet_string {
    #       uint16_t length;
    #       uint8_t data[length];
    #   };
    #   keyblock {
    #       uint16_t type;
    #       counted_octet_string key;
    #   };

    def __init__(self):
        self.entries = []
        self.filename = None

    @classmethod
    def load(cls, path):
        keytab = cls()
        keytab.read(path)
        return keytab

    def _read(self, f, fmt):
        """Read and unpack one or more fields."""
        size = struct.calcsize(fmt)
        data = f.read(size)
        if len(data) != size:
            raise IOError("Failed to read keytab data.")
        return struct.unpack(fmt, data) # returns a tuple

    def _write(self, f, fmt, *args):
        """Pack and write one or more fields."""
        data = struct.pack(fmt, *args)
        f.write(data)

    def _read_string(self, f):
        """Read and unpack a string."""
        length = self._read(f, "!H")
        string, = self._read(f, "%ds" % length)
        return string

    def _write_string(self, f, s):
        """Pack and write a string."""
        self._write(f, "!H%ds" % (len(s)), len(s), s)

    def _read_entry(self, f, version):
        numc, = self._read(f, "!h")
        if version == 0x0501:
            numc -= 1
        realm = self._read_string(f)
        comps = []
        for i in xrange(0, numc):
            comps.append(self._read_string(f))
        if version != 0x0501:
            name_type, = self._read(f, "!L")
        timestamp, = self._read(f, "!L")
        vno8, = self._read(f, "!B")
        eno, = self._read(f, "!H")
        key = self._read_string(f)
        principal = "%s@%s" % ('/'.join(comps), realm)
        entry = {
            'realm': realm,
            'components': comps,
            'principal': principal,
            'timestamp': timestamp,
            'kvno': vno8,
            'eno': eno,
            'enctype': KRB_ENCTYPES.get(eno, 'unknown'),
            'key': key,
        }
        return entry

    def _write_entry(self, f, k):
        # first pass: predicate the size to be written
        fmt = ["!"]
        fmt.append("h")  # number of components
        fmt.append("H%ds" % (len(k['realm'])))
        for c in k['components']:
            fmt.append("H%ds" % (len(c))) # c[i]
        fmt.append("L") # KRB_NT_PRINCIPAL
        fmt.append("L") # timestamp
        fmt.append("B") # kvno
        fmt.append("H") # eno
        fmt.append("H%ds" % len(k['key'])) # key
        size = struct.calcsize("".join(fmt))
        # second pass: write the data
        self._write(f, "!l", size)
        self._write(f, "!h", len(k['components']))
        self._write_string(f, k['realm'])
        for c in k['components']:
            self._write_string(f, c)
        self._write(f, "!L", Keytab._KRB_NT_PRINCIPAL)
        self._write(f, "!L", k['timestamp'])
        self._write(f, "!B", k['kvno'])
        self._write(f, "!H", k['eno'])
        self._write_string(f, k['key'])

    def _kdf(self, secret, size):
        """For test purposes only."""
        h = hashlib.sha256()
        if secret:
            h.update(secret)
        else:
            h.update(os.urandom(size))
        d = h.digest()
        return d.ljust(size)[:size]

    def _generate_des_key(self, secret):
        """Generate a random DES key with correct parity bits."""
        keybytes = bytearray(self._kdf(secret, 8))
        key = bytearray(0)
        for i in keybytes:
            b = bin(i & 0xfe)
            nb = len(b.split('1'))
            # nb is one more than the number of bits set
            key.append(int(b, 2) + (nb % 2))
        return bytes(key)

    def read(self, path):
        """Read the entries from a keytab file.

        Read the data from the keytab file directly. This avoid requiring
        the external klist utility on systems where akimpersonate is
        used instead of kinit.
        """
        logger.debug("reading keytab %s", path)
        self.filename = path
        try:
            stat = os.stat(path)
        except OSError as e:
            if e.errno == os.errno.ENOENT:
                raise NotFoundKeytab("File {0} not found.".format(path))
            else:
                raise e
        file_size = stat.st_size
        with open(path, 'r') as f:
            offset = struct.calcsize("!h")
            try:
                version, = self._read(f, "!h")
            except IOError:
                raise InvalidKeytab("File {0} is not keytab.".format(path))
            if version != 0x0502 and version != 0x051:  # magic
                raise InvalidKeytab("File {0} is not keytab.".format(path))
            while offset < file_size:
                f.seek(offset)
                size, = self._read(f, "!l") # record size, not including this field
                entry = self._read_entry(f, version)
                self.entries.append(entry)
                # Calculate next record location.
                record_size = struct.calcsize("!l") + size
                offset += record_size
        return self

    def add_fake_key(self, principal, enctype='aes256-cts-hmac-sha1-96', secret=None):
        """Create a synthetic key for testing with akimpersonate."""
        logger.debug("Adding fake key for principal '%s', enctype='%s'.", principal, enctype)
        timestamp = int(time.time())
        kvno = 1
        comps,realm = _split_principal(principal)
        if isinstance(enctype, int):
            eno = enctype
        else:
            eno = _lookup_eno(enctype)
        if _is_des_eno(eno):
            key = self._generate_des_key(secret)
        elif eno == 17:
            key = self._kdf(secret, 16)
        elif eno == 18:
            key = self._kdf(secret, 32)
        else:
            raise AssertionError("Unsupported enctype for fake keytab: eno=%d" % eno)
        entry = {
           'realm': realm,
           'components': comps,
           'principal': principal,
           'timestamp': timestamp,
           'kvno': kvno,
           'eno': eno,
           'enctype': KRB_ENCTYPES.get(eno, 'unknown'),
           'key': key,
        }
        self.entries.append(entry)
        return self

    def write(self, path):
        logger.debug("Writing keytab %s", path)
        mkdirp(os.path.dirname(path))
        with open(path, 'w') as f:
            self._write(f, '!h', self._KRB_KEYTAB_MAGIC)
            for k in self.entries:
                self._write_entry(f, k)
        self.filename = path
        return self

    def get_kvno(self, principal):
        """Find the largest kvno for the given principal.

        None is returned if no matches are found."""
        kvno = None
        for e in self.entries:
            if e['principal'] == principal:
                if e['kvno'] > kvno:
                    kvno = e['kvno']
        return kvno

    def get_entries(self, principal, kvno=None):
        """Return the list of entries for the given principal and kvno.

        If knvo is not given, then the entries with the largest kvno are
        returned. None is returned if no matches are found."""
        if kvno is None:
            kvno = self.get_kvno(principal)
            if kvno is None:
                return None
        entries = []
        for e in self.entries:
            if e['principal'] == principal and e['kvno'] == kvno:
                entries.append(e)
        return entries

    def is_des(self, principal, kvno=None):
        """Returns true if a DES enctype is present for the given principal.

        Raises an exception if no keys are found for the given principal
        or if a mix of DES and non-DES kes are present."""
        has_des = False
        has_non_des = False
        entries = self.get_entries(principal, kvno)
        if entries is None:
            raise AssertionError("No keytab entries found for %s!" % (principal))
        for e in entries:
            if _is_des_eno(e['eno']):
                has_des = True
            else:
                has_non_des = True
        # Do not allow a mix of des and non-des keys!
        if has_des and has_non_des:
            raise AssertionError("DES and non-DES entries found for %s!" % (principal))
        return has_des

    def service_principals(self):
        """Return the list of afs service principals in this keytab.

        Only modern service principal formats are supported, that is
        the service principals must be in the form afs/<cellname>@<REALMNAME>."""
        sp = set()
        for e in self.entries:
            if e['principal'].startswith("afs/"):
                sp.add(e['principal'])
        return list(sp)

    def _guess_key_format(self, principal):
        """Attempt to detect the key file format for this keytab on this server.

        Returns one of: 'transarc', 'rxkad-k5', or 'extended'.  'rxkad-k5' is
        assumed if the key is not DES and asetkey does not support the extended
        key file format."""
        logger.info("Attempting to detect key format.")
        if self.is_des(principal):
            kformat = 'transarc'
        elif _check_for_extended_keyfile_support():
            kformat = 'extended'
        else:
            # Assume rxkad-k5 is supported for this non-DES service key.
            kformat = 'rxkad-k5'
        logger.info("Detected key format is '%s'." % (kformat))
        return kformat

    def _set_service_key_transarc(self, principal):
        """Set the DES service key."""
        keytab = self.filename
        kvno = "%d" % self.get_kvno(principal)
        asetkey('add', kvno, keytab, principal)

    def _set_service_key_rxkad_k5(self, confdir, principal):
        """Set the non-DES service key for rxkad-k5 support."""
        # Verify this keytab does not contain a DES key for the principal.
        if self.is_des(principal):
            raise AssertionError("Cannot use DES key for rxkad-k5!")
        if not os.path.isdir(confdir):
            raise ValueError("Cannot find server config directory '%s'." % (confdir))
        confdir = os.path.abspath(confdir)
        src = self.filename
        dst = os.path.join(confdir, 'rxkad.keytab')
        if os.path.exists(dst):
            logger.info("Removing old rxkad-k5 file: %s" % (dst))
            os.remove(dst)
        logger.info("Copying rxkad-k5 keytab %s to %s.", src, dst)
        shutil.copyfile(src, dst)
        os.chmod(dst, 0600)

    def _set_service_key_extended(self, principal):
        """Set the non-DES service key."""
        keytab = self.filename
        for e in self.get_entries(principal):
            kvno = "%d" % e['kvno']
            eno = "%d" % e['eno']
            asetkey('add', 'rxkad_krb5', kvno, eno, keytab, principal)

    def _get_service_principal(self, cell, realm):
        if cell is not None:
            if realm is None:
                realm = cell.upper()
            principal = "afs/%s@%s" % (cell, realm)
        else:
            sp = self.service_principals()
            if len(sp) == 1:
                principal = sp[0]
            elif len(sp) == 0:
                raise AssertionError("The keytab '%s' does not contain any afs service principals." % (self.filename))
            elif len(sp) > 1:
                raise AssertionError("The keytab '%s' has more than one afs service principal; "
                                     "Specify which cell and realm should be used." % (self.filename))
            else:
                raise AssertionError("Internal error")
        return principal

    def set_service_key(self, cell, realm, kformat, confdir, dryrun=False, **kwargs):
        """Set the service key given in a keytab file."""
        if self.filename is None:
            raise AssertionError("Must read or write keytab file first.""")

        principal = self._get_service_principal(cell, realm)
        logger.debug("Using service principal '%s'.", principal)
        kvno = self.get_kvno(principal)
        if kvno is None:
            raise ValueError("Principal '%s' not found in keytab %s" % (principal, self.filename))
        enctypes = [e['enctype'] for e in self.get_entries(principal)]

        if not kformat:
            kformat = 'detect'
        if kformat == 'detect':
            kformat = self._guess_key_format(principal)

        if not dryrun:
            msg = "Setting key"
        else:
            msg = "Skipping set key for dryrun"
        logger.info("%s; keytab '%s', principal '%s', kvno '%s', enctypes '%s', format '%s'",
                    msg, self.filename, principal, kvno, ", ".join(enctypes), kformat)

        if kformat == 'transarc':
            if not dryrun:
                self._set_service_key_transarc(principal)
        elif kformat == 'rxkad-k5':
            if not dryrun:
                self._set_service_key_rxkad_k5(confdir, principal)
        elif kformat == 'extended':
            if not dryrun:
                self._set_service_key_extended(principal)
        else:
            raise ValueError("Unknown key file format name: %s" % (kformat))

    def akimpersonate(self, user=None, cell=None, realm=None, **kwargs):
        if self.filename is None:
            raise AssertionError("Must read or write keytab file first.")
        if user is None:
            user = DEFAULT_USER
        user = user.replace('.', '/')  # convert k4-style separators (used by afs) to k5-style
        sp = self._get_service_principal(cell, realm) # lookup the cell/realm if needed
        if self.get_kvno(sp) is None:
            raise AssertionError("No service principal found!")
        cell,realm = sp.replace('afs/', '').split('@')
        principal = "%s@%s" % (user, realm)

        output = aklog('-d', '-c', cell, '-k', realm, '-principal', principal, '-keytab', self.filename)
        for line in output.splitlines():
            logger.info(line)

def create(cell='robotest', realm=None, keytab='/tmp/afs.keytab',
           enctype=None, secret=None, **kwargs):
    """Create a fake keytab file.

    cell: afs cell name
    realm: kerberos realm name
    keytab: destination file path
    enctype: encyption type
    secret: key data
    """
    if realm is None:
        realm = cell.upper()
    principal = "afs/{0}@{1}".format(cell, realm)
    k = Keytab()
    k.add_fake_key(principal, enctype=enctype, secret=secret)
    k.write(keytab)

def destroy(keytab='/tmp/afs.keytab', force=False, **kwargs):
    """Remove a keytab file.

    keytab: path of the keytab file to be removed
    force:  delete even if the file is not a valid keytab
    """
    try:
        k = Keytab().load(keytab)
        logger.debug("removing keytab file %s", k.filename)
        os.remove(k.filename)
    except NotFoundKeytab:
        logger.debug("keytab not found %s", keytab)
    except InvalidKeytab:
        if not force:
            logger.info("skipping removal of non-keytab file %s", keytab)
        else:
            logger.debug("removing invalid keytab file %s (forced)", keytab)
            os.remove(keytab)

def setkey(keytab='/tmp/afs.keytab', cell=None, realm=None,
        kformat=None, confdir=AFS_CONF_DIR, **kwargs):
    """Set a service key from a keytab file.

    keytab: path of the keytab file
    cell: cellname
    realm: kerberos realm name
    kformat: keyfile format
    confdir: server config file path
    """
    k = Keytab.load(keytab)
    k.set_service_key(cell, realm, kformat, confdir, **kwargs)




"""Misc functions"""

def flatten(x):
    """Flatten a small list of lists.

    Example:
        flatten( [[1, 2], [3]] ) => [1, 2, 3]
    """
    if not isinstance(x, (list, tuple)):
        return x # skip non-interables, including None
    y = []
    for e in x:
        if isinstance(e, (list, tuple)):
            y.extend(flatten(e))
        else:
            y.append(e)
    return y

def lists2dict(options):
    """ Helper to convert a list of lists of name=value strings to a dict."""
    if isinstance(options, dict):
        return options
    if options is None:
        options = []
    if not isinstance(options, (list, tuple)):
        options = [options]
    options = flatten(options)
    names = {}
    for o in options:
        name,value = o.split('=', 1)
        names[name.strip()] = value.strip()
    return names

def trim(s, x):
    """Trim trailing x from s."""
    if s.endswith(x):
        s = s.rsplit(x,1)[0]
    return s

def uniq(x):
    """Remove dupes from a small list, preserving order."""
    y = []
    for item in x:
        if not item in y:
            y.append(item)
    return y

def flipbase64_encode(n):
    """Encode an integer to a flip-base64, a base64 variant for AFS namei."""
    index = '+=0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz'
    base64 = [index[n & 0x3f]]
    n >>= 6
    while n:
        base64.append(index[n & 0x3f])
        n >>= 6
    return ''.join(base64)



# Copyright (c) 2018 Sine Nomine Associates
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THE SOFTWARE IS PROVIDED 'AS IS' AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

"""Helper to install and remove OpenAFS RPM packages via yum."""

import logging

from afsutil.rpm import RpmInstaller
from afsutil.system import sh
from afsutil.install import Installer

logger = logging.getLogger(__name__)

def yum(*args):
    """Helper to run the yum command."""
    return sh('yum', *args, quiet=False)

class YumInstaller(RpmInstaller):
    """Install and remove OpenAFS packages with yum."""

    def __init__(self, dir=None, **kwargs):
        Installer.__init__(self, **kwargs)
        self.pkgdir = dir  # Defer checks until needed.
        self.packages = None
        self.installed = None

    def install(self):
        """Install RPM packages."""
        self.pre_install()
        if not (self.do_server or self.do_client):
            raise AssertionError("Expected client and/or server component.")
        packages = []
        packages.append('openafs')
        packages.append('openafs-krb5')
        packages.append('openafs-docs')
        if self.do_server:
            packages.append('openafs-server')
        if self.do_client:
            packages.append('openafs-client')
            packages.append('kmod-openafs')
        yum('-y', 'install', *packages)
        self.post_install()

    def remove(self):
        """Remove OpenAFS RPM packages."""
        # Remove all packages by default. Optionaly remove just the server or
        # client packages. If removing just one component and the other is not
        # present, then also remove the common packages.
        self.pre_remove()
        installed = self.find_installed().keys()
        packages = [] # names of packages to be removed
        if self.do_server and self.do_client:
            packages = installed # remove everything
        elif self.do_server:
            if 'openafs-client' not in installed:
                packages = installed              # remove common too
            elif 'openafs-server' in installed:
                packages = ['openafs-server']
        elif self.do_client:
            if 'openafs-server' not in installed:
                packages = installed              # remove common too
            elif 'openafs-client' in installed:
                packages = ['openafs-client', 'kmod-openafs']
        if packages:
            logger.info("removing %s" % " ".join(packages))
            yum('-y', 'remove', *packages)
        self.post_remove()



# Copyright (c) 2014-2017 Sine Nomine Associates
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THE SOFTWARE IS PROVIDED 'AS IS' AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

"""Solaris specific utilities."""

import logging
import os
import re

from afsutil.system import common as _mod
CommandMissing = _mod.CommandMissing
CommandFailed = _mod.CommandFailed
cat = _mod.cat
directory_should_exist = _mod.directory_should_exist
directory_should_not_exist = _mod.directory_should_not_exist
file_should_exist = _mod.file_should_exist
mkdirp = _mod.mkdirp
nproc = _mod.nproc
path_join = _mod.path_join
sh = _mod.sh
symlink = _mod.symlink
touch = _mod.touch
which = _mod.which

logger = logging.getLogger(__name__)

def get_running():
    """Get a set of running processes."""
    ps = which('/usr/bin/ps') # avoid the old BSD variant
    lines = sh(ps, '-e', '-f', quiet=True)
    # The first line of the `ps' output is a header line which is
    # used to find the data field columns.
    column = lines[0].index('CMD')
    procs = set()
    for line in lines[1:]:
        cmd_line = line[column:]
        command = cmd_line.split()[0]
        procs.add(os.path.basename(command))
    return procs

def is_running(program):
    """Returns true if program is running."""
    return program in get_running()

def afs_mountpoint():
    mountpoint = None
    pattern = r'(/.\S+) on AFS'
    mount = which('mount', extra_paths=['/bin', '/sbin', '/usr/sbin'])
    output = sh(mount, quiet=True)
    for line in output:
        found = re.search(pattern, line)
        if found:
            mountpoint = found.group(1)
    return mountpoint

def is_afs_mounted():
    """Returns true if afs is mounted."""
    return afs_mountpoint() is not None

def afs_umount():
    """Attempt to unmount afs, if mounted."""
    afs = afs_mountpoint()
    if afs:
        umount = which('umount', extra_paths=['/bin', '/sbin', '/usr/sbin'])
        sh(umount, afs)

def network_interfaces():
    """Return list of non-loopback network interfaces."""
    try:
        command = which('ipadm')
        args = ('show-addr', '-p', '-o', 'STATE,ADDR')
        pattern = r'ok:(\d+\.\d+\.\d+\.\d+)'
    except CommandMissing:
        # Fall back to old command on old solaris releases.
        command = which('/usr/sbin/ifconfig')
        args = ('-a')
        pattern = r'inet (\d+\.\d+\.\d+\.\d+)'
    addrs = []
    output = sh(command, *args)
    for line in output:
        match = re.match(pattern, line)
        if match:
            addr = match.group(1)
            if not addr.startswith("127."):
                addrs.append(addr)
    return addrs

def is_loaded(kmod):
    """Returns the list of currently loaded kernel modules."""
    output = sh('/usr/sbin/modinfo', '-w')
    for line in output[1:]: # skip header line
        # Fields are: Id Loadaddr Size Info Rev Module (Name)
        m = re.match(r'\s*(\d+)\s+\S+\s+\S+\s+\S+\s+\d+\s+(\S+)', line)
        if not m:
            raise AssertionError("Unexpected modinfo output: %s" % (line))
        mid = m.group(1)  # We will need the id to remove the module.
        mname = m.group(2)
        if kmod == mname:
            return mid
    return 0

def detect_gfind():
    return which('gfind', extra_paths=['/opt/csw/bin'])

def tar(tarball, source_path, tar=None):
    if tar is None:
        tar = 'gtar'
    sh(tar, 'czf', tarball, source_path, quiet=True)

def untar(tarball, chdir=None, tar=None):
    if tar is None:
        tar = 'gtar'
    savedir = None
    if chdir:
        savedir = os.getcwd()
        os.chdir(chdir)
    try:
        sh(tar, 'xzf', tarball, quiet=True)
    finally:
        if savedir:
            os.chdir(savedir)

def _so_symlinks(path):
    """Create shared lib symlinks."""
    if not os.path.isdir(path):
        assert AssertionError("Failed to make so symlinks: path '%s' is not a directory.", path)
    for dirent in os.listdir(path):
        fname = os.path.join(path, dirent)
        if os.path.isdir(fname) or os.path.islink(fname):
            continue
        m = re.match(r'(.+\.so)\.(\d+)\.(\d+)\.(\d+)$', fname)
        if m:
            so,x,y,z = m.groups()
            symlink(fname, "%s.%s.%s" % (so, x, y))
            symlink(fname, "%s.%s" % (so, x))
            symlink(fname, so)

def configure_dynamic_linker(path):
    if not os.path.isdir(path):
        raise AssertionError("Failed to configure dynamic linker: path %s not found." % (path))
    _so_symlinks(path)
    sh('/usr/bin/crle', '-u', '-l', path)
    sh('/usr/bin/crle', '-64', '-u', '-l', path)

def unload_module():
    module_id = is_loaded('afs')
    if module_id != 0:
        sh('modunload', '-i', module_id)



# Copyright (c) 2014-2016 Sine Nomine Associates
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THE SOFTWARE IS PROVIDED 'AS IS' AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

"""Install and remove Transarc-style OpenAFS distributions."""

import logging
import os
import shutil
import glob
import tempfile
import urllib2

import afsutil
from afsutil.install import Installer, \
                            copy_files, remove_file, remove_files

from afsutil.system import sh, directory_should_exist, \
                           configure_dynamic_linker, \
                           is_loaded, is_running, mkdirp, path_join, \
                           untar

logger = logging.getLogger(__name__)

#
# Transarc style install paths.
#
AFS_CONF_DIR = "/usr/afs/etc"
AFS_KERNEL_DIR = "/usr/vice/etc"
AFS_SRV_BIN_DIR =  "/usr/afs/bin"
AFS_SRV_SBIN_DIR = "/usr/afs/bin"
AFS_SRV_LIBEXEC_DIR = "/usr/afs/bin"
AFS_SRV_LIB_DIR = "/usr/afs/lib"
AFS_DB_DIR = "/usr/afs/db"
AFS_LOGS_DIR = "/usr/afs/logs"
AFS_LOCAL_DIR = "/usr/afs/local"
AFS_BACKUP_DIR = "/usr/afs/backup"
AFS_BOS_CONFIG_DIR = "/usr/afs/local"
AFS_DATA_DIR = "/usr/vice/etc"
AFS_CACHE_DIR = "/usr/vice/cache"
AFS_MOUNT_DIR = "/afs"
AFS_WS_DIR = "/usr/afsws"
SYSCONFIG = "/etc/sysconfig"

#
# Fully qualified AFS command paths for Transarc style installs.
#
AKLOG = "/usr/afsws/bin/aklog"
ASETKEY = "/usr/afs/bin/asetkey"
BOS = "/usr/afs/bin/bos"
FS = "/usr/afs/bin/fs"
PTS = "/usr/afs/bin/pts"
RXDEBUG = "/usr/afsws/etc/rxdebug"
TOKENS = "/usr/afsws/bin/tokens"
UDEBUG = "/usr/afs/bin/udebug"
UNLOG = "/usr/afsws/bin/unlog"
VOS = "/usr/afs/bin/vos"

class TransarcClientSetup(object):
    """Transarc client specific setup functions."""

    def install_afsd(self, afsd):
        """Install the afsd file."""
        # Common for linux an solaris.
        src = afsd
        dst = AFS_KERNEL_DIR
        logger.info("Installing afsd from '%s' to '%s'.", src, dst)
        mkdirp(dst)
        shutil.copy2(src, dst)
        os.chmod(dst, 0755)

class LinuxClientSetup(TransarcClientSetup):
    """Linux specific setup functions."""

    def install_init_script(self, dest, afsd_options):
        """Install a client init script on linux.

        Does not configure the system to run the init script automatically on
        reboot.
        """
        # Install the init script.
        mkdirp("/var/lock/subsys/")
        text = afsutil.resources['openafs-client-linux.init']
        dst = "/etc/init.d/openafs-client"
        logger.info("Installing client init script from to '%s'.", dst)
        with open(dst, 'w') as f:
            f.write(text)
        os.chmod(dst, 0755)
        # Set afsd options.
        dst = path_join(SYSCONFIG, "openafs-client")
        logger.info("Writing afsd startup options to file '%s'." % (dst))
        mkdirp(os.path.dirname(dst))
        with open(dst, 'w') as f:
            f.write('AFSD_OPTIONS="%s"\n' % (afsd_options))

    def _install_openafs_ko(self, kmod):
        """Install the openafs.ko file and run depmod."""
        release = os.uname()[2]
        src = kmod
        dst = path_join("/lib/modules", release, "extra/openafs/openafs.ko")
        logger.info("Installing kernel module from '%s' to '%s'.", src, dst)
        mkdirp(os.path.dirname(dst))
        shutil.copy2(src, dst)
        sh('/sbin/depmod', '-a')

    def install_driver(self, dest, force=False):
        #
        # If available, install the openafs.ko to a path loadable by modprobe.
        # Use the openafs configure option --with-linux-kernel-packaging to build
        # openafs.ko instead of libafs${version}.ko
        #
        # Note: libafs${version}.ko was already installed by copying all the files from
        #       dest/root.client/usr/vice/etc to /usr/vice/etc.
        #
        release = os.uname()[2]
        kmod = path_join(dest, "root.client", "lib/modules", release, "extra/openafs/openafs.ko")
        if os.path.exists(kmod):
            self._install_openafs_ko(kmod)

    def remove_driver(self):
        release = os.uname()[2]
        src = path_join("/lib/modules", release, "extra/openafs/openafs.ko")
        if os.path.exists(src):
            remove_file(src)

    def install_kmod(self, kmod):
        """Install the kernel module by filename."""
        if os.path.basename(kmod) == 'openafs.ko':
            self._install_openafs_ko(kmod)
        elif os.path.basename(kmod).startswith('libafs'):
            src = kmod
            dst = path_join(AFS_KERNEL_DIR, 'modload')
            logger.info("Installing kernel module from '%s' to '%s'.", src, dst)
            shutil.copy2(src, dst)
        else:
            raise ValueError("Unknown kernel module name: %s" % (kmod))

class SolarisClientSetup(TransarcClientSetup):
    """Solaris specific setup functions."""

    def install_init_script(self, dest, afsd_options):
        """Install a client init script on solaris.

        Does not configure the system to run the init script automatically on
        reboot.  Changes the init script to avoid starting the bosserver
        by default."""
        osrel = os.uname()[2]
        text = afsutil.resources['openafs-client-solaris-%s.init' % (osrel)]
        dst = "/etc/init.d/openafs-client"
        logger.info("Installing client init script to '%s'.", dst)
        with open(dst, 'w') as f:
            f.write(text)
        os.chmod(dst, 0755)
        # Set afsd options.
        config = '/usr/vice/etc/config'
        dst = path_join(config, 'afsd.options')
        logger.info("Writing afsd startup options to file '%s'." % (dst))
        mkdirp(os.path.dirname(dst))
        with open(dst, 'w') as f:
            f.write(afsd_options) # Not sourced.

    def _afs_driver(self):
        """Return the name of the afs driver for the current platform."""
        osrel = os.uname()[2]
        if osrel == '5.10':
            return self._afs_driver_510()
        elif osrel == '5.11':
            return self._afs_driver_511()
        else:
            raise AssertionError("Unsupported operating system: %s" % (osrel))

    def _afs_driver_510(self):
        output = sh('/bin/isalist')[0]
        if 'amd64' in output:
            driver = '/kernel/fs/amd64/afs'
        elif 'sparcv9' in output:
            driver = '/kernel/fs/sparcv9/afs'
        else:
            driver = '/kernel/fs/afs'
        return driver

    def _afs_driver_511(self):
        output = sh('/bin/isalist')[0]
        if 'amd64' in output:
            driver = '/kernel/drv/amd64/afs'
        elif 'sparcv9' in output:
            driver = '/kernel/drv/sparcv9/afs'
        else:
            driver = '/kernel/drv/afs'
        return driver

    def install_driver(self, dest):
        afs = "libafs64.o"
        kmod = path_join(dest, "root.client", AFS_KERNEL_DIR, "modload", afs)
        self.install_kmod(kmod)

    def remove_driver(self):
        remove_file(self._afs_driver())
        remove_file('/kernel/drv/afs.conf')

    def install_kmod(self, kmod):
        """Install the kernel module by filename."""
        src = kmod
        dst = self._afs_driver()
        logger.info("Installing kernel driver from '%s' to '%s'.", src, dst)
        shutil.copy2(src, dst)

class TransarcInstaller(Installer):
    """Install a Transarc-style distribution of OpenAFS on linux or solaris."""

    def __init__(self, **kwargs):
        """Initialize the Transarc-style installer.

        """
        Installer.__init__(self, **kwargs)
        self.bins = kwargs.get('dir', None)
        self.tmpfile = None
        self.tmpdir = None
        self.url = None
        self.tarball = None
        self.dest = None
        uname = os.uname()[0]
        if uname == "Linux":
            self.client_setup = LinuxClientSetup()
        elif uname == "SunOS":
            self.client_setup = SolarisClientSetup()
        else:
            raise AssertionError("Unsupported operating system: %s" % (uname))
        self.installed = {'libs':False, 'client':False, 'server':False, 'ws':False}

    def _detect_dest(self):
        user = os.getenv('SUDO_USER') # The user who invoked sudo.
        patterns = ["./*/dest"]
        if user:
            patterns.append("/home/{user}/openafs/*/dest".format(**locals()))
            patterns.append("/home/{user}/src/openafs/*/dest".format(**locals()))
        for pattern in patterns:
            logger.debug("Searching for dest in '%s'", pattern)
            paths = glob.glob(pattern)
            if len(paths) > 1:
                raise AssertionError("Unable to find dest directory: too many sysnames in '%s'!" % (pattern))
            if len(paths) == 1:
                path = os.path.abspath(paths[0])
                logger.debug("Found dest: '%s'", path)
                return path
        raise AssertionError("Unable to find dest directory.")

    def _check_dest(self, dest):
        """Verify the dest directory looks sane."""
        directory_should_exist(dest)
        directory_should_exist(path_join(dest, 'root.server'))
        directory_should_exist(path_join(dest, 'root.client'))
        directory_should_exist(path_join(dest, 'lib'))

    def _install_shared_libs(self, src, dst):
        """Install the shared libraries."""
        if self.installed['libs']:
            logger.debug("Skipping shared libs install; already done.")
        else:
            copy_files(src, dst, symlinks=False, force=self.force)
            configure_dynamic_linker(dst)
            self.installed['libs'] = True

    def _install_server_rc(self):
        """Install a server init script.

        The Transarc-style distributions provide a single init script which
        optionally starts the bosserver. This is an historical relic left over
        from the pre-namei fileserver days. Instead, install a simple, separate
        init script to start and stop the bosserver.

        Does not configure the system to run the init script automatically on
        reboot.
        """
        text = afsutil.resources['openafs-server.init']
        dst = "/etc/init.d/openafs-server"
        if os.path.exists(dst) and not self.force:
            raise AssertionError("Refusing to overwrite '%s'.", dst)
        with open(dst, 'w') as f:
            f.write(text)
        os.chmod(dst, 0755)
        mkdirp("/var/lock/subsys/")  # needed by the init script
        # Set the bosserver command line options.
        dst = path_join(SYSCONFIG, "openafs-server")
        logger.info("Writing bosserver startup options to file '%s'." % (dst))
        mkdirp(os.path.dirname(dst))
        with open(dst, 'w') as f:
            f.write('BOSSERVER_OPTIONS="%s"\n' % (self.options.get('bosserver', '')))

    def _install_workstation_binaries(self):
        """Install workstation binaries from a Transarc-style distribution."""
        if self.installed['ws']:
            logger.debug("Skipping workstation files install; already done.")
        else:
            for d in ('bin', 'etc', 'include', 'man'):
                src = path_join(self.dest, d)
                dst = path_join(AFS_WS_DIR, d)
                copy_files(src, dst, force=self.force)
            self.installed['ws'] = True
        src = path_join(self.dest, 'lib')
        dst = path_join(AFS_WS_DIR, 'lib')
        self._install_shared_libs(src, dst)

    def _install_server(self):
        """Install server binaries."""
        logger.info("Installing server binaries")
        src = path_join(self.dest, "root.server", AFS_SRV_BIN_DIR)
        copy_files(src, AFS_SRV_BIN_DIR, force=self.force)
        self._install_shared_libs(path_join(self.dest, 'lib'), AFS_SRV_LIB_DIR)
        self._install_workstation_binaries() # rxdebug is in the ws directory.
        self._install_server_rc()
        self.installed['server'] = True
        logger.info("Servers installed.")

    def _install_client(self):
        """Install client binaries."""
        logger.info("Installing client binaries")
        src = path_join(self.dest, "root.client", AFS_KERNEL_DIR)
        copy_files(src, AFS_KERNEL_DIR, force=self.force) # also installs libafs.ko
        self._install_workstation_binaries() # including libs, unless already installed
        self.client_setup.install_driver(self.dest)
        self.client_setup.install_init_script(self.dest, self.options.get('afsd', ''))
        mkdirp(AFS_MOUNT_DIR)
        mkdirp(AFS_CACHE_DIR)
        self.installed['client'] = True
        logger.info("Client installed.")

    def _remove_server(self):
        """Remove the server binaries."""
        if is_running('bosserver'):
            raise AssertionError("Refusing to remove: bosserver is running.")
        remove_files("/usr/afs/bin/")
        remove_files("/usr/afs/lib/")
        remove_file("/etc/init.d/openafs-server")

    def _remove_client(self):
        """Remove the client binaries."""
        if is_running('afsd'):
            raise AssertionError("Refusing to remove: afsd is running.")
        if is_loaded('libafs'):
            raise AssertionError("Refusing to remove: libafs is loaded.")
        # is /afs mounted?
        remove_files("/usr/vice/etc/")
        remove_files("/usr/afsws/")
        remove_file("/etc/init.d/openafs-client")
        remove_file("/etc/sysconfig/afs")
        self.client_setup.remove_driver()

    def install(self):
        """Install Transarc-style distribution OpenAFS binaries.

        self.bins specifies the bins to be installed.
        * None: Install the bins under <sysname>/dest in the current directory.
        * a directory path:  Install the bins found in the given dest directory.
        * a path to a tarball: Untar the tarball in a temporary directory and
          install the bins found under <sysname>/dest.
        * url path: Download the tarball file at the given url, untar the tarball,
          and install the bins found under <sysname>/dest.
        """
        def _untar(tarball):
            self.tmpdir = tempfile.mkdtemp()
            logger.info("Untarring %s into %s", self.bins, self.tmpdir)
            untar(tarball, chdir=self.tmpdir)
            return glob.glob("%s/*/dest" % (self.tmpdir))[0] # must have a dest dir.

        def _download(url):
            BLOCK_SIZE = 4 * 1024
            rsp = urllib2.urlopen(url)
            (fh, self.tmpfile) = tempfile.mkstemp(suffix='.tar.gz')
            logger.info("Downloading %s to %s", url, self.tmpfile)
            with open(self.tmpfile, 'wb') as tarball:
                block = rsp.read(BLOCK_SIZE)
                while block:
                    tarball.write(block)
                    block = rsp.read(BLOCK_SIZE)
            os.close(fh)
            return self.tmpfile

        if self.bins is None:
            self.dest = self._detect_dest()
        elif os.path.isdir(self.bins):
            self.dest = self.bins
        elif (self.bins.startswith('http://') or self.bins.startswith('https://')) and self.bins.endswith('.tar.gz'):
            self.dest = _untar(_download(self.bins))
        elif os.path.isfile(self.bins) and self.bins.endswith('.tar.gz'):
            self.dest = _untar(self.bins)
        else:
            raise AssertionError("Unrecognized path to installation files: %s" % (self.bins))
        self._check_dest(self.dest)
        self.pre_install()
        if self.do_server:
            self._install_server()
        if self.do_client:
            self._install_client()
        self.post_install()
        if self.tmpfile:
            logger.info("Removing tmp file %s", self.tmpfile)
            os.remove(self.tmpfile)
        if self.tmpdir:
            logger.info("Removing tmp dir %s", self.tmpdir)
            shutil.rmtree(self.tmpdir)

    def remove(self):
        """Remove Transarc-style distribution OpenAFS binaries."""
        self.pre_remove()
        if self.do_client:
            self._remove_client()
        if self.do_server:
            self._remove_server()
        self.post_remove()

#
# Test Driver
#
class _Test(object):
    def __init__(self, dest):
        self.dest = dest

    def test_install(self):
        i = TransarcInstaller(dest=self.dest, force=True)
        i.install()

    def test_remove(self):
        i = TransarcInstaller(dest=self.dest, purge=True)
        i.remove()

    def test_install_server(self):
        i = TransarcInstaller(dest=self.dest, force=True, components=['server'])
        i.install()

    def test_remove_server(self):
        i = TransarcInstaller(dest=self.dest, force=True, components=['server'])
        i.remove()

    def test_install_client(self):
        i = TransarcInstaller(dest=self.dest, force=True, components=['client'])
        i.install()

    def test_remove_client(self):
        i = TransarcInstaller(dest=self.dest, force=True, components=['client'])
        i.remove()

    def test(self):
        logging.basicConfig(level=logging.DEBUG)
        self.test_remove()
        self.test_install_server()
        self.test_install_client()
        self.test_remove_server() # leaves common packages
        self.test_remove_client() # removes common packages



# Copyright (c) 2014-2016 Sine Nomine Associates
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THE SOFTWARE IS PROVIDED 'AS IS' AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

"""Helper to install and remove OpenAFS RPM packages."""

import logging
import os
import glob

from afsutil.system import sh
from afsutil.install import Installer

logger = logging.getLogger(__name__)

def rpm(*args):
    """Helper to run the rpm command."""
    return sh('rpm', *args, quiet=True)

class RpmInstaller(Installer):
    """Helper to install and remove OpenAFS RPMs."""

    def __init__(self, dir=None, **kwargs):
        Installer.__init__(self, **kwargs)
        self.pkgdir = dir  # Defer checks until needed.
        self.packages = None
        self.installed = None

    def find_installed(self):
        """Find the installed openafs packages on this system."""
        # We get all of packages and check the names here since the rpm
        # command on this system could be old and not support wildcards.
        self.installed = {}
        output = rpm('--query', '--all',
                     '--queryformat', '%{NAME} %{VERSION} %{RELEASE} %{ARCH}\\n')
        for line in output:
            name,version,release,arch = line.split()
            if name.startswith('kmod-openafs') or name.startswith('openafs'):
                self.installed[name] = {
                    'version': version,
                    'release': release,
                    'arch': arch}
        return self.installed

    def find_packages(self):
        """Find the OpenAFS rpm package files.

        Search for the RPM files to be installed. The search order is:
        1. The 'dir' named argument given to the RpmInstaller(), if one.
        2. The current working directory.
        3. The ./packages/rpmbuild/RPMS/<arch> directory. (This is where
           the `afsutil package` command places the built rpms.)"""
        # Older versions of this tried to parse the filenames to get package
        # attributes. This version uses rpm query in instead, which seems to
        # to be less error prone.  We build a list instead of a dict, since
        # there could be more than one rpm per package name, e.g.,
        # kmod_openafs*.
        if self.packages is not None:
            return self.packages  # Lazy evaluated.
        # Setup search paths.
        packages = []
        search_paths = []
        cwd = os.getcwd()
        arch = os.uname()[4] # e.g. x86_64
        if self.pkgdir:
            search_paths.append(self.pkgdir)
        search_paths.append(cwd)
        search_paths.append(os.path.join(cwd, "packages/rpmbuild/RPMS/%s" % (arch)))
        for sp in search_paths:
            files = glob.glob(os.path.join(sp, '*.rpm'))
            for f in files:
                if f.endswith('.src.rpm'): # Skip the source rpm, if present.
                    continue
                output = rpm('--query', '--package', f,
                             '--queryformat', '%{NAME} %{VERSION} %{RELEASE} %{ARCH}\\n')
                name,version,release,arch = output[0].split()
                packages.append({'file': f, 'name': name, 'version': version,
                                'release': release, 'arch': arch})
            if len(packages) != 0:
                break
        if len(packages) == 0:
            raise AssertionError("No packages found in paths: '%s'" % ":".join(search_paths))
        self.packages = packages
        return self.packages

    def find_package(self, name):
        """Get a package rpm file by package name."""
        # To simplify, assume we will not mix different versions of userspace
        # rpms in the package directory.
        packages = self.find_packages()
        found = []
        for p in packages:
            if p['name'] == name:
                found.append(p)
        if len(found) == 0:
            raise AssertionError("Package '%s' not found in '%s'." % (name, self.pkgdir))
        if len(found) > 1:
            raise AssertionError("Multiple packages for '%s' found in '%s'." % (name, self.pkgdir))
        return found[0]

    def find_kmod(self):
        """Find the kmod package matching the kernel version on this system."""
        # This functions handles rhel/centos openafs kmod naming.  The release
        # name in the kmod contains the kernel version number without the
        # system arch.  Convert the underscore in the package release to a dash
        # to match the kernel's version format.
        kernel = os.uname()[2]  # e.g. '2.6.32-431.29.2.el6.x86_64'
        logger.info("Searching for module package for kernel version %s.", kernel)
        packages = self.find_packages()
        for p in packages:
            if p['name'].startswith('kmod-openafs'):
                arch = p['arch']  #  e.g., 'x86_64'
                release = p['release'].replace('_', '-') # e.g. '1.2.6.32_431.29.2.el6'
                release = ".".join((release, arch))
                if release.find(kernel) != -1:
                    return p  # kernel version found
        raise AssertionError("Cannot find module package for kernel version %s." % (kernel,))

    def install(self):
        """Install RPM packages."""
        self.pre_install()
        if not (self.do_server or self.do_client):
            raise AssertionError("Expected client and/or server component.")
        packages = []
        packages.append(self.find_package('openafs'))
        packages.append(self.find_package('openafs-krb5'))
        packages.append(self.find_package('openafs-docs'))
        if self.do_server:
            packages.append(self.find_package('openafs-server'))
        if self.do_client:
            packages.append(self.find_package('openafs-client'))
            packages.append(self.find_kmod())
        files = []
        for p in packages:
            file = p['file']
            files.append(file)
            logger.info("Installing %s.", file)
        if len(files) == 0:
            raise AssertionError("No rpm files found to install.")
        rpm('-v', '--install', '--replacefiles', '--replacepkgs', *files)
        self.post_install()

    def remove(self):
        """Remove OpenAFS RPM packages."""
        # Remove all packages by default. Optionaly remove just the server or
        # client packages. If removing just one component and the other is not
        # present, then also remove the common packages.
        self.pre_remove()
        installed = self.find_installed().keys()
        packages = [] # names of packages to be removed
        if self.do_server and self.do_client:
            packages = installed # remove everything
        elif self.do_server:
            if 'openafs-client' not in installed:
                packages = installed              # remove common too
            elif 'openafs-server' in installed:
                packages = ['openafs-server']
        elif self.do_client:
            if 'openafs-server' not in installed:
                packages = installed              # remove common too
            elif 'openafs-client' in installed:
                packages = ['openafs-client', 'kmod-openafs']
        if packages:
            logger.info("removing %s" % " ".join(packages))
            rpm('-v', '--erase', *packages)
        self.post_remove()



# Copyright (c) 2014-2017 Sine Nomine Associates
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THE SOFTWARE IS PROVIDED 'AS IS' AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

"""Linux specific utilities."""

import logging
import os
import re

from afsutil.system import common as _mod
CommandMissing = _mod.CommandMissing
CommandFailed = _mod.CommandFailed
cat = _mod.cat
directory_should_exist = _mod.directory_should_exist
directory_should_not_exist = _mod.directory_should_not_exist
file_should_exist = _mod.file_should_exist
mkdirp = _mod.mkdirp
nproc = _mod.nproc
path_join = _mod.path_join
sh = _mod.sh
symlink = _mod.symlink
touch = _mod.touch
which = _mod.which

logger = logging.getLogger(__name__)

def get_running():
    """Get a set of running processes."""
    ps = which('ps')
    lines = sh(ps, '-e', '-f', quiet=True)
    # The first line of the `ps' output is a header line which is
    # used to find the data field columns.
    column = lines[0].index('CMD')
    procs = set()
    for line in lines[1:]:
        cmd_line = line[column:]
        if cmd_line[0] == '[':  # skip linux threads
            continue
        command = cmd_line.split()[0]
        procs.add(os.path.basename(command))
    return procs

def is_running(program):
    """Returns true if program is running."""
    return program in get_running()

def afs_mountpoint():
    mountpoint = None
    pattern = r'^AFS on (/.\S+)'
    mount = which('mount', extra_paths=['/bin', '/sbin', '/usr/sbin'])
    output = sh(mount, quiet=True)
    for line in output:
        found = re.search(pattern, line)
        if found:
            mountpoint = found.group(1)
    return mountpoint

def is_afs_mounted():
    """Returns true if afs is mounted."""
    return afs_mountpoint() is not None

def afs_umount():
    """Attempt to unmount afs, if mounted."""
    afs = afs_mountpoint()
    if afs:
        umount = which('umount', extra_paths=['/bin', '/sbin', '/usr/sbin'])
        sh(umount, afs)

def network_interfaces():
    """Return list of non-loopback network interfaces."""
    addrs = []
    output = sh('/sbin/ip', '-oneline', '-family', 'inet', 'addr', 'show', quiet=True)
    for line in output:
        match = re.search(r'inet (\d+\.\d+\.\d+\.\d+)', line)
        if match:
            addr = match.group(1)
            if not addr.startswith("127."):
                addrs.append(addr)
    logger.debug("Found network interfaces: %s", ",".join(addrs))
    return addrs

def is_loaded(kmod):
    with open("/proc/modules", "r") as f:
        for line in f.readlines():
            if kmod == line.split()[0]:
                return True
    return False

def configure_dynamic_linker(path):
    """Configure the dynamic linker with ldconfig.

    Add a path to the ld configuration file for the OpenAFS shared
    libraries and run ldconfig to update the dynamic linker."""
    conf = '/etc/ld.so.conf.d/openafs.conf'
    paths = set()
    paths.add(path)
    if os.path.exists(conf):
        with open(conf, 'r') as f:
            for line in f.readlines():
                line = line.strip()
                if line.startswith("#") or line == "":
                    continue
                paths.add(line)
    with open(conf, 'w') as f:
        logger.debug("Writing %s", conf)
        for path in paths:
            f.write("%s\n" % path)
    sh('/sbin/ldconfig')

def unload_module():
    output = sh('/sbin/lsmod')
    for line in output:
        kmods = re.findall(r'^(libafs|openafs)\s', line)
        for kmod in kmods:
            sh('rmmod', kmod)

def detect_gfind():
    return which('find')

def tar(tarball, source_path, tar=None):
    if tar is None:
        tar = 'tar'
    sh(tar, 'czf', tarball, source_path, quiet=True)

def untar(tarball, chdir=None, tar=None):
    if tar is None:
        tar = 'tar'
    savedir = None
    if chdir:
        savedir = os.getcwd()
        os.chdir(chdir)
    try:
        sh(tar, 'xzf', tarball, quiet=True)
    finally:
        if savedir:
            os.chdir(savedir)



"""Utility classes to build, install, and setup OpenAFS."""

from afsutil.__version__ import VERSION as __version__
from afsutil.__resources__ import resources
__version__   # hush lint
resources     # hush lint



try:
    from setuptools import setup
except ImportError:
    # Fallback to the standard library distutils.
    from distutils.core import setup

exec(open('afsutil/__version__.py').read()) # sets VERSION

description='Utilities to setup OpenAFS clients and servers'
try:
    long_description = open('README.rst').read()
except:
    # Fallback if readme is not found.
    long_description = description

setup(
    name='afsutil',
    version=VERSION,
    description=description,
    long_description=long_description,
    author='Michael Meffie',
    author_email='mmeffie@sinenomine.net',
    url='https://github.com/openafs-contrib/afsutil',
    license='BSD',
    packages=[
        'afsutil',
        'afsutil.system',
    ],
    scripts=[
        'bin/afsutil',
    ],
    include_package_data=True,
    test_suite='test',
    zip_safe=False,
    options = {
        'bdist_rpm':{
            'post_install': 'packaging/rpm/post_install.sh',
            'post_uninstall': 'packaging/rpm/post_uninstall.sh',
        },
    },
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: POSIX :: Linux',
        'Operating System :: POSIX :: SunOS/Solaris',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Topic :: Utilities',
    ],
)
